#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import logging
import re
import commands
import definitions
import text_mixer
from file import File
from execute import ExternalCommand

class ScsiHost(object):
    def __init__(self, scsi_host):
        self.host_name = scsi_host
        self.host_name_number = scsi_host.replace('host','')
        self.get_wwn()
        self.type = 'unknown type'
        self.read_proc_scsi_scsi()

    def read_proc_scsi_scsi(self):
        self.proc_scsi_scsi_file = File('/proc/scsi/scsi')
        try:
            self.proc_scsi_scsi_data = self.proc_scsi_scsi_file.read_file().splitlines()
        except IOError:
            logging.error('failed to read /proc/scsi/scsi')
            self.proc_scsi_scsi_data = ''

    def scan(self):
        logging.info("Scanning for new LUNs for %s" %(self.host_name))
        self.remove_bad_luns()
        self.issue_lip()
        self.add_new_luns()
    
    def issue_lip(self):
        issue_lip_command = ExternalCommand(''' echo 1 > /sys/class/fc_host/%s/issue_lip''' % (self.host_name))
        issue_lip_command.execute()
        if issue_lip_command.return_value == 256:
            logging.debug('failed to echo 1 to issue_lip, but that is alright - the adapter does not support it')

    def send_scan_to_scsi_host_scan(self):
        driver_scan_command = ExternalCommand('''echo '- - -' > /sys/class/scsi_host/%s/scan''' % (self.host_name))
        driver_scan_command.execute()

    def add_new_luns(self):
        self.send_scan_to_scsi_host_scan()
        self.scan_by_add_single_device()

    def scan_by_add_single_device(self):
        for hct in self.get_targets_from_lun_zero():
            logging.debug('scanning target %s for new LUNs' % hct)
            for lun in range(0, definitions.MAX_LUN_SCAN): 
                add_single_device = ExternalCommand(\
                    '''echo 'scsi add-single-device %s %s' > /proc/scsi/scsi''' % (hct, lun))
                add_single_device.execute()

    def remove_bad_luns(self):
        devices_list = []
        devices_list.extend(self.get_removed_devices())
        for device in devices_list:
                logging.debug('removing single device %s' % device)
                remove_single_device_command = ExternalCommand(\
                    '''echo 'scsi remove-single-device %s' > /proc/scsi/scsi''' % (device))
                remove_single_device_command.execute()

    def get_removed_devices(self):
        hctl_list = self.get_hctl_list_for_this_host()
        removed_hctl_list = []
        for hctl in hctl_list:
            if self.need_to_remove_device(hctl):
                removed_hctl_list.append(hctl)
        return removed_hctl_list

    def need_to_remove_device(self, hctl):
        device_node = self.map_hctl_to_device(hctl)
        if not self.is_device_responding(device_node):
            return True
        if text_mixer.is_lun_zero(hctl):
            if not self.read_capacity(device_node):
                return True
        return False

    def read_capacity(self, device_node):
        sg_readcap = ExternalCommand('sg_readcap -vvv %s' % device_node)
        sg_readcap.execute()
        if sg_readcap.contains('failed'):
            return False
        return True

    def get_hctl_list_for_this_host(self):
        hctl_list = []
        for line in self.proc_scsi_scsi_data:
            if line.__contains__('Host: scsi%s' % self.host_name_number):
                (host, channel, target, lun) = text_mixer.extract_hctl_from_line(line)
                logging.debug('found an XIV LUN: %s %s %s %s' % (host, channel, target, lun))
                hctl_list.append('%s %s %s %s' % (host, channel, target, lun))
        return hctl_list

    def get_targets_from_lun_zero(self):
        hct_list = []
        for line in self.proc_scsi_scsi_data:
            if line.__contains__('Host: scsi%s' % self.host_name_number):
                (host, channel, target, lun) = text_mixer.extract_hctl_from_line(line)
                if lun == '0' and not hct_list.__contains__('%s %s %s' % (host, channel, target)):
                    logging.debug('found an XIV target: %s %s %s %s' % (host, channel, target, lun))
                    hct_list.append('%s %s %s' % (host, channel, target))
        return hct_list
        
    def map_hctl_to_device(self, hctl):
        sg_map = ExternalCommand('sg_map -x')
        sg_map.execute()
        for line in sg_map.output.splitlines():
            if line.__contains__('Strange'):
                logging.debug('sg_map reported strange errors')
                return ''
            if line.__contains__(hctl):
                return line.split(' ').pop()
        return ''

    def is_device_responding(self, device):
        sg_turs = ExternalCommand('sg_turs %s' % (device))
        sg_turs.execute()
        if sg_turs.is_error_free():
            return True
        return False

    def get_wwn(self):
        try:
            port_name_file = open('/sys/class/fc_host/%s/port_name' % (self.host_name), 'ro')
            self.wwn = re.sub(r'(..)', r'\1:', port_name_file.readline().strip().split('x')[1]).strip(':')
            port_name_file.close()
        except IOError:
            logging.debug('Cannot retreive WWN for %s' % (self.host_name))
            self.wwn = ''

