#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import logging
import os
import time
from file import File
from shutil import copyfile

import definitions
from execute import ExternalCommand
from service import GenericServiceManager
from hostattachment.exceptions import HostAttachmentError

class IIscsidConf(object):
    def __init__(self):
        raise NotImplementedError("Abstract Class")
#        self.log_used_classes()
#    def log_used_classes(self):
#        self.logger.log_a_class(ConfigFile)
#        self.logger.log_a_class(ConfigFileWriter)
#        self.logger.log_a_class(ConfigFileReader)

    def set_dest_filename(self):
        raise NotImplementedError()

    def load_config(self):
        raise NotImplementedError()

    def is_configured(self):
        if self.dest_config_missing():
            logging.error('file %s is empty or missing' % self.dest_filename)
            return False
        if self.configuration_mismatch():
            logging.error('configuration file %s is not valid' \
                % self.dest_filename)
            return False
        return True

    def source_config_missing(self):
        return self.source_config.is_empty()

    def dest_config_missing(self):
        return self.dest_config.is_empty()

    def configuration_mismatch(self):
        for (name, value) in self.source_config.settings.iteritems():
            match = False
            if self.dest_config.settings.has_key(name):
                if self.dest_config.settings[name] == value:
                    match = True
            if not match:
                logging.debug('iSCSI configuration file is missing '+ \
                    'the %s parameter' % name)
                return True
        return False

    def configure(self):
        if self.dest_config.is_empty():
            if not self.override_existing_config():
                raise HostAttachmentError("iscsid.conf write error")
            logging.info('%s configured successfully' % self.dest_filename)
        else:
            try:
                self.dest_config.backup()
            except Exception, e:
                raise HostAttachmentError('Failed to backup existing iSCSI '+ \
                    'configuration, aborting')
            if not self.write_merged_config():
                raise HostAttachmentError("iscsid.conf write error")
            logging.info('%s configured successfully' % self.dest_filename)

    def override_existing_config(self):
        if not self.source_config.write_to_file(self.dest_filename):
            return False
        return True

    def write_merged_config(self):
        if not self.merge_configuration():
            logging.error('failed to merge configuration')
            return False
        if not self.dest_config.write_to_file():
            logging.error('failed to write to file %s' % self.dest_filename)
            return False
        return True

    def merge_configuration(self):
        for (source_name, source_value) in self.source_config.settings.iteritems():
            self.dest_config.add_setting(source_name, source_value)
        return True

class IIscsiGeneral(object):
    def __init__(self):
#        self.log_used_classes()
        raise NotImplementedError("Abstract class")

#    def log_used_classes(self):
#        self.logger.log_a_class(ConnectionManager)
#        self.logger.log_a_class(InitiatorName)

    def get_configuration_task_list(self):
        tasks = []

        if not self.initiator.is_configured():
            tasks.append("Initiator not configured")
        if not self.iscsid_conf.is_configured():
            tasks.append("iscsi.conf not configured")
        if not self.service.is_configured():
            tasks.append("iscsi service not configured")
        if not self.iscsiadm.is_configured():
            tasks.append("iscsiadm not configured")

        return tasks

    def configure(self):
        self.initiator.configure()
        self.service.configure()
        self.iscsid_conf.configure()

    def connect(self, ip_address):
        return self.iscsiadm.connect(ip_address)

    def disconnect(self, ip_address):
        return self.iscsiadm.disconnect(ip_address)

    def get_iqn(self):
        return self.initiator.get_iqn()

    def print_all(self):
        print self.initiator.get_iqn()
        self.iscsiadm.print_all()

    def scan(self):
        self.iscsiadm.refresh_connections()


class ConfigFileReader(object):
    def __init__(self, filepath):
        self.filepath = filepath
        self.config_string = ''
        self.config_lines = []
        self.settings = []

    def read(self):
        config_file = File(self.filepath)
        try:
            self.config_string = config_file.read_file()
        except IOError:
            logging.debug('failed to read %s' % self.filepath)
            return False
        self.strip_comments_and_newlines_and_tabs()
        return True

    def strip_comments_and_newlines_and_tabs(self):
        for line in self.config_string.splitlines():
            stripped_line = line.strip().replace('\t', ' ')
            if not stripped_line == '' and not stripped_line.startswith('#'):
                logging.debug('found a non-empty an not-commented line: %s' % stripped_line)
                self.config_lines.append(stripped_line)


class ConfigFileWriter(object):
    def __init__(self, filepath, settings = {}, discovery_addresses=[]):
        self.filepath = filepath
        self.settings = settings
        self.discovery_addresses = discovery_addresses

    def write_to_file(self):
        self.lines = []
        self.append_header()
        self.append_settings()
        self.append_discovery_addresses()
        try:
            file_object = open(self.filepath, 'w+')
            for line in self.lines:
                file_object.write('%s\n' % line)
            file_object.flush()
            file_object.close()
        except IOError:
            logging.error('failed to open %s for writing' % self.filepath)
            return False
        return True

    def append_header(self):
        self.lines.append('# iSCSI Initiator Configuration File')
        self.lines.append('# Version %s' % definitions.PACKAGE_VERSION)
        (tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, tm_yday, tm_isdst) = time.localtime()
        self.lines.append('# Generated by IBM-XIV Host Attachment at %s-%s-%s %s:%s:%s' % (tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec))
        self.lines.append('# Do not manually modify this file!')
        self.lines.append('')

    def append_settings(self):
        for (name, value) in self.settings.iteritems():
            self.lines.append('%s=%s' % (name, value))

    def append_discovery_addresses(self):
        for address in self.discovery_addresses:
            self.lines.append('DiscoveryAddress=%s' % address)


class IConfigFile(object):
    def __init__(self, filepath):
        raise NotImplementedError("Abstract class")

    def read(self):
        if self.reader.read():
            logging.debug('successfully read %s' % self.filepath)
            self.analyze()
        else:
            logging.debug('could not read %s' % self.filepath)

    def analyze(self):
        for line in self.reader.config_lines:
            (name, value) = self.split_n_strip_line(line)
            logging.debug('found %s=%s' % (name, value))
            self.add_setting(name, value)

    def split_n_strip_line(self, line):
        (name, value) = line.split('=', 1)
        return (name.strip(), value.strip())

    def add_setting(self, name, value):
        self.settings[name] = value

    def write_to_file(self, filepath=''):
        if filepath == '':
            filepath = self.filepath 
        writer = ConfigFileWriter(filepath, self.settings)
        return writer.write_to_file()

    def is_empty(self):
        if len(self.settings) == 0:
            return True
        return False

    def backup(self):
        backup_file = "%s%s%s.%s" % (self.backup_path, os.path.sep,
            os.path.basename(self.filepath), time.strftime("%Y%m%d%H%m%S"))
        copyfile(self.filepath, backup_file)

class IInitiatorName(object):
    def __init__(self):
        raise NotImplementedError("Abstract class")

    def set_dest_filename(self):
        self.dest_filename = '/etc/initiatorname.iscsi'
 
    def get_initiator_name(self):
        self.name = None
        if not self.read_initiator_file():
            return False
        if not self.file_contains_iqn():
            return False
        self.extract_iqn()

    def read_initiator_file(self):
        try:
            self.initiator_file_content = self.initiator_file.read_file()
        except IOError:
            return False
        return True

    def file_contains_iqn(self):
        if not self.initiator_file_content.__contains__('InitiatorName'):
            logging.error('%s does not contain InitiatorName' % \
                (self.dest_filename))
            return False
        return True

    def extract_iqn(self):
        for line in self.initiator_file_content.splitlines():
            if line.strip().startswith('InitiatorName='):
                self.name = line.strip().split('=')[1]

    def set_initiator_name(self, iqn):
        self.initiator_file.write_file(data='InitiatorName=%s' % iqn)

    def configure(self):
        if self.name is None:
            generate = ExternalCommand('iscsi-iname')
            generate.execute()
            if not generate.is_error_free():
                raise HostAttachmentError(\
                    "Failed configuring iSCSI initiator")
            self.set_initiator_name(generate.output)

    def is_configured(self):
        if self.name is None:
            logging.error('InitiatorName is not found')
            return False
        else:
            return True

    def get_iqn(self):
        return self.name
