#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import logging

import definitions
from config_file import ConfigFile
from device import Device
from reader import ConfigFileReader
from writer import ConfigFileWriter
from hostattachment.exceptions import HostAttachmentError

class MultiPathConf(object):
    def __init__(self, backup_path):
#        self.log_used_classes()
        self.backup_path = backup_path
        self.source_file = definitions.MULTIPATH_CONF_SOURCE
        self.dest_filename = '/etc/multipath.conf'

        self.load_config()

    def log_used_classes(self):
        self.logger.log_a_class(ConfigFile)
        self.logger.log_a_class(Device)
        self.logger.log_a_class(ConfigFileWriter)
        self.logger.log_a_class(ConfigFileReader)

    def load_config(self):
        self.source_config = ConfigFile(self.source_file, self.backup_path)
        self.dest_config = ConfigFile(self.dest_filename, self.backup_path)
        self.source_config.read()
        self.dest_config.read()

    def is_configured(self):
        if self.source_config_missing():
            logging.error('file %s is empty or missing' % self.source_file)
            return False
        if self.dest_config_missing():
            logging.error('file %s is empty or missing' % self.dest_filename)
            return False
        if self.default_config_mismatch():
            logging.warn('defaults in %s are not valid' % self.dest_filename)
        if self.blacklist_contains_xiv():
            logging.error('%s blocked XIV devices' % self.dest_filename)
            return False
        if self.whitelist_missing_xiv():
            logging.error('%s does not have XIV devices as whitelisted' % self.dest_filename)
            return False
        return True

    def source_config_missing(self):
        return self.source_config.is_empty()

    def dest_config_missing(self):
        return self.dest_config.is_empty()

    def default_config_mismatch(self):
        return self.dest_config.defaults != self.source_config.defaults

    def blacklist_contains_xiv(self):
        return self.blacklist_devnodes_blacklist_sd() or self.blacklist_devices_contains_xiv()

    def blacklist_devnodes_blacklist_sd(self):
        for devnode in self.dest_config.blacklisted_devnodes:
            if devnode.startswith('''"*'''):
                return True
            if devnode.startswith('''"^*'''):
                return True
            if devnode.__contains__('''"sd'''):
                return True
        return False

    def blacklist_devices_contains_xiv(self):
        for source_device in self.source_config.whitelisted_devices:
            for dest_device in self.dest_config.blacklisted_devices:
                for key in source_device.identifiers:
                    try:
                        if dest_device.identifiers[key] == source_device.identifiers[key]:
                            logging.error('XIV devices are blacklisted in %s' % self.dest_filename)
                            return True
                    except KeyError:
                        pass
        return False

    def whitelist_missing_xiv(self):
        for source_device in self.source_config.whitelisted_devices:
            xiv_device_missing = False
            for dest_device in self.dest_config.whitelisted_devices:
                if dest_device.identifiers == source_device.identifiers:
                    if dest_device.settings == source_device.settings:
                        logging.debug('XIV device matched from source to destination multipath.conf file')
                    else:
                        logging.error('XIV device listed in multipath.conf, but its settings are invalid')
                        return True
            if xiv_device_missing:
                logging.error('XIV devices are missing from existing multipath.conf file')
                return True
        return False

    def configure(self):
        if self.dest_config.is_empty():
            self.override_existing_config()
            logging.info('%s configured successfully' % self.dest_filename)
        try:
            self.dest_config.backup()
        except Exception, e:
            raise HostAttachmentError('failed to backup existing '+\
                'multipath.conf, aborting')
        self.write_merged_config()
        logging.info('%s configured successfully' % self.dest_filename)

    def override_existing_config(self):
        if not self.source_config.write_to_file(self.dest_filename):
            raise HostAttachmentError("Failed writing file %s" %\
            (self.dest_filename))

    def write_merged_config(self):
        self.merge_whitelists()
        self.merge_defaults()
        if not self.source_config.write_to_file(self.dest_filename):
            raise HostAttachmentError('failed to write to file %s' %\
                self.dest_filename)

    def merge_whitelists(self):
        for dest_device in self.dest_config.whitelisted_devices:
            add_device = True
            for source_device in self.source_config.whitelisted_devices:
                for key in source_device.identifiers:
                    try:
                        if dest_device.identifiers[key] == source_device.identifiers[key]:
                            logging.debug('XIV device found in the existing multipath.conf, will not merge it')
                            add_device = False
                    except KeyError:
                        pass
            if add_device:
                logging.info('non-XIV device found in the existing multipath.conf, adding it to the new configuration')
                for key in dest_device.identifiers:
                    logging.debug('device identifier: %s=%s' % (key, dest_device.identifiers[key]))
                if not self.source_config.add_device_to_whitelist(dest_device):
                    logging.error('failed to add the device to the configuration')
                    raise HostAttachmentError('Failed to match whitelists')

    def merge_defaults(self):
        for key in self.dest_config.defaults:
            if not self.source_config.defaults.__contains__(key):
                self.source_config.defaults[key] = self.dest_config.defaults[key]
