#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
#
#-------------------------------------------------------------------------------------------------

import logging
import os
from shutil import copyfile
import time

from execute import ExternalCommand
from device import Device
from reader import ConfigFileReader
from writer import ConfigFileWriter

class ConfigFile(object):
    def __init__(self, filepath, backup_path):
        self.filepath = filepath
        self.backup_path = backup_path
        self.instance_objects()

    def instance_objects(self):
        self.defaults_lines = []
        self.blacklist_lines = []
        self.whitelist_lines = []
        self.defaults = {}
        self.blacklisted_devnodes = []
        self.blacklisted_devices = []
        self.whitelisted_devices = []
        self.reader = ConfigFileReader(self.filepath)

    def read(self):
        if self.reader.read():
            logging.debug('successfully read %s' % self.filepath)
            self.analyze()
        else:
            logging.debug('could not read %s' % self.filepath)

    def analyze(self):
        self.analyze_defaults()
        self.analyze_blacklist()
        self.analyze_whitelist()

    def analyze_defaults(self):
        self.defaults_lines = self.reader.get_block('defaults')
        defaults_line = False
        indentation_level = 1
        for line in self.defaults_lines:
            logging.debug('(i=%s) line = %s' % (str(indentation_level), line))
            if line == '}':
                indentation_level-=1;
                if indentation_level == 1:
                    defaults_line = False
            if defaults_line and indentation_level == 2:
                splits = line.split(' ',1)
                self.defaults[splits[0].strip()] = splits[1].strip()
            if line.startswith('defaults') and indentation_level == 1:
                defaults_line = True
            if line.endswith('{'):
                indentation_level+=1
        for key in self.defaults:
            logging.debug('found default parameter: %s=%s' % (key, self.defaults[key]))

    def analyze_blacklist(self):
        self.blacklist_lines = self.reader.get_block('blacklist')
        self.get_blacklisted_devnodes()
        self.get_blacklisted_devices()

    def analyze_whitelist(self):
        self.whitelist_lines = self.reader.get_block('devices')
        self.get_whitelisted_devices()

    def get_blacklisted_devnodes(self):
        indentation_level = 1
        for line in self.blacklist_lines:
            logging.debug('(i=%s) line = %s' % (str(indentation_level), line))
            if line == '}':
                indentation_level-=1
            elif line.__contains__('devnode') and indentation_level == 2:
                devnode = line.split(' ',1)[1].strip()
                logging.debug('found a blacklist devnode: %s' % devnode)
                self.blacklisted_devnodes.append(devnode)
            elif line.endswith('{'):
                indentation_level+=1

    def get_blacklisted_devices(self):
        self.blacklisted_devices = self.reader.get_devices_from_block(self.blacklist_lines)

    def get_whitelisted_devices(self):
        self.whitelisted_devices = self.reader.get_devices_from_block(self.whitelist_lines)

    def write_to_file(self, filepath=''):
        if filepath == '':
            filepath = self.filepath 
        writer = ConfigFileWriter(filepath, self.defaults, self.blacklisted_devnodes, self.blacklisted_devices, self.whitelisted_devices)
        return writer.write_to_file()

    def is_empty(self):
        if self.reader.config_string == '':
            return True
        return False

    def add_device_to_whitelist(self, device):
        self.whitelisted_devices.append(device)
        return True
    
    def backup(self):
        backup_file = "%s%s%s.%s" % (self.backup_path, os.path.sep,
            os.path.basename(self.filepath), time.strftime("%Y%m%d%H%m%S"))
        copyfile(self.filepath, backup_file)
