#!/usr/bin/env xpyv

#-------------------------------------------------------------------------------------------------
# Copyright (c) International Business Machines Corp., 2009
#
# IBM XIV Storage System Software
#
# Licensed materials provided under the terms of the IBM International Program  
# License Agreement. See the Software
# licensing materials that came with the IBM XIV Storage System for terms and
# conditions.
# 
#-------------------------------------------------------------------------------------------------   

from os import path, environ, system
from subprocess import PIPE, Popen
from shutil import copyfile
from tempfile import mktemp

from interfaces import IXIVDiag
from collector.collectable.files import SimpleFileCollectable as FileItem
from collector.collectable.commands import SimpleCommandCollectable as CommandItem
from collector.collectable.directories import SimpleDirectoryCollectable as DirectoryItem

from win32evtlog import BackupEventLog, OpenEventLog, CloseEventLog
from win32api import GetComputerName
from xpyv import Wow64FsRedirection,project_root_path

HOST_ATTACHMENT_DIR = path.abspath(path.join(project_root_path.find(), path.pardir, path.pardir, path.pardir))
DEFAULTDIR = path.join(environ['SystemRoot'], 'Temp')
CLUSTER_2008_EXE_PATH = path.join(environ['SystemRoot'], 'System32', 'cluster.exe')
CLUSTER_2003_LOG_PATH = path.join(environ['SystemRoot'], 'Cluster', 'cluster.log')
CLUSTER_2008_LOG_PATH = path.join(environ['SystemRoot'], 'System32', 'winevt', 'Logs', 'ClusterLog.etl')
CMD_PATH = path.join(environ['SystemRoot'], 'System32', 'cmd.exe')
DISKPART_PATH = path.join(environ['SystemRoot'], 'System32', 'diskpart.exe')
REG_PATH = path.join(environ['SystemRoot'], 'System32', 'reg.exe')
MSINFO32_PATH = path.join(environ['ProgramFiles'].strip(" (x86)"), 'Common files', 'Microsoft shared', 'MSinfo', 'msinfo32.exe')
MPIO_LIST_PATH = path.join(HOST_ATTACHMENT_DIR, 'drivers', 'mpio', 'mpiolist.exe')
XIV_DEVLIST_PATH = path.join(HOST_ATTACHMENT_DIR, 'bin', 'xiv_devlist')
MINIDUMP_PATH = path.join(environ['SystemRoot'], 'Minidump')      
MEMORYDUMP_PATH = path.join(environ['SystemRoot'], 'MEMORY.DMP')
SETUPAPI_NT5_PATH = path.join(environ['SystemRoot'], 'setupapi.log')
SETUPAPI_NT6_PATH = path.join(environ['SystemRoot'], 'setupapi.log')

class XIVDiag(IXIVDiag):
    def __init__(self):
        self.get_computername()
        IXIVDiag.__init__(self, DEFAULTDIR)
        Wow64FsRedirection.disable()

    def get_computername(self):
        self.computername = GetComputerName()

    def add_collectables(self):
        self.add_msinfo32_collectables()
        self.add_eventlogs_collectables()
        self.add_clusterlog_collectables()
        self.add_mpiolist_collectables()
        self.add_diskpart_collectables()
        self.add_registry_collectables()
        self.add_services_collectables()
        self.add_kernel_dumps_collectables()
        self.add_setupapi_collectables()
        self.add_reg_hardware_subtree()
        self.add_xiv_devlist()
        
    def add_msinfo32_collectables(self):
        temp_file = mktemp(prefix='msinfo32_')
        temp_file = '%s.nfo' % temp_file
        self.add_msinfo32_command_collectable(temp_file)
        self.add_msinfo32_file_collectable(temp_file)

    def add_msinfo32_command_collectable(self, temp_file):
        self.add_collectable_item(CommandItem( \
            "System Information (1/2)", [MSINFO32_PATH, '/nfo', temp_file], timeout=180))

    def add_msinfo32_file_collectable(self, temp_file):
        self.add_collectable_item(FileItem("System Information (2/2)", self.zip_file, temp_file))

    def add_eventlogs_collectables(self):
        for eventlog in ('System', 'Application'):
            eventlog_tempfile = self.prepare_eventlog(eventlog)
            self.add_collectable_item(FileItem("%s Event Log" % eventlog, self.zip_file, eventlog_tempfile))

    def prepare_eventlog(self, eventlog):
        eventlog_tempfile = mktemp(prefix='%s_eventlog' % eventlog)
        try:
            eventlog_object = OpenEventLog(self.computername, eventlog)
            BackupEventLog(eventlog_object, eventlog_tempfile)
            CloseEventLog(eventlog_object)
        except:
            pass
        return eventlog_tempfile

    def add_clusterlog_collectables(self):
        self.add_collectable_item(FileItem("Cluster (2003) Log", self.zip_file, CLUSTER_2003_LOG_PATH)) 
        self.add_collectable_item(CommandItem("Cluster (2008) Log Generator", [CLUSTER_2008_EXE_PATH, 'log', '/generate']))
        for i in range(1,6):
            self.add_collectable_item(FileItem("Cluster (2008) Logs (%s/5)" % i, self.zip_file, '%s.00%s' % (CLUSTER_2008_LOG_PATH, i)))

    def add_mpiolist_collectables(self):
        for i in range(0,6):
            self.add_collectable_item(CommandItem("MultiPath I/O Information (%s/6)" % str(i+1), [MPIO_LIST_PATH, str(i)]))
    
    def add_diskpart_collectables(self):
        self.add_collectable_item(CommandItem('DISKPART: List Disk', [CMD_PATH, '/c echo LIST DISK | %s' % DISKPART_PATH]))
        self.add_collectable_item(CommandItem('DISKPART: List Volume', [CMD_PATH, '/c echo LIST VOLUME | %s' % DISKPART_PATH]))

    def add_registry_collectables(self):
        self.add_collectable_item(CommandItem('Installed HotFixes', \
            [REG_PATH, 'query', '''HKLM\SOFTWARE\Microsoft\Windows NT\CurrentVersion\HotFix''', '/s']))
        self.add_collectable_item(CommandItem('DSMXIV Configuration', \
            [REG_PATH, 'query', '''HKLM\SYSTEM\CurrentControlSet\Services\dsmxiv''', '/s']))

    def add_services_collectables(self):
        self.add_collectable_item(CommandItem('Services Information', ['sc', 'query']))
    
    def add_kernel_dumps_collectables(self):
        minidump_temp_file, memorydump_temp_file = self.copy_kernel_dumps_to_tempdir()
        self.add_collectable_item(FileItem('Windows Minidump', self.zip_file, minidump_temp_file))
        self.add_collectable_item(FileItem('Windows Memory Dump', self.zip_file, memorydump_temp_file))

    def add_setupapi_collectables(self):
        self.add_collectable_item(FileItem('Windows Setup API (1/2)', self.zip_file, SETUPAPI_NT5_PATH))
        self.add_collectable_item(FileItem('Windows Setup API (2/2)', self.zip_file, SETUPAPI_NT6_PATH))

    def add_reg_hardware_subtree(self):
        self.add_collectable_item(CommandItem('Hardware Registry Subtree', \
            [REG_PATH, 'query', '''HKLM\HARDWARE''', '/s']))

    def copy_kernel_dumps_to_tempdir(self):
        print ('INFO: Copying Memory dumps to temporary directory... ').ljust(70, ' ')
        output_tempfile = mktemp(prefix='output_')
        minidump_tempfile = mktemp(prefix='minidump_')
        memorydump_tempfile = mktemp(prefix='memorydump_')
        output = system('copy /y %s %s > %s' % (MINIDUMP_PATH, minidump_tempfile, output_tempfile))
        output = system('copy /y %s %s > %s' % (MEMORYDUMP_PATH, memorydump_tempfile, output_tempfile))
        print 'DONE'.rjust(70, ' ')
        return (minidump_tempfile, memorydump_tempfile)

    def add_xiv_devlist(self):
        self.add_collectable_item(CommandItem('xiv_devlist' , \
            [XIV_DEVLIST_PATH]))


